using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.VisualStudio.Modeling.Validation;

namespace ImTryin.ClassesLanguage
{
    [ValidationState(ValidationState.Enabled)]
    public partial class ClassesModel
    {
        [ValidationMethod(ValidationCategories.Open | ValidationCategories.Save | ValidationCategories.Menu)]
        public void CheckForUniqueNames(ValidationContext context)
        {
            Dictionary<string, List<Type>> typesByName = new Dictionary<string, List<Type>>();
            foreach (Type type in Types)
            {
                string fullTypeName = ((type.Namespace.Length > 0) ? type.Namespace + "." : "") + type.Name;
                if (!typesByName.ContainsKey(fullTypeName))
                    typesByName[fullTypeName] = new List<Type>();
                typesByName[fullTypeName].Add(type);
            }
            foreach (String name in typesByName.Keys)
            {
                List<Type> types = typesByName[name];
                if (types.Count > 1)
                    context.LogError("Classes can not have same names inside the same namespace", "SameNames",
                                     types.ToArray());
            }
        }
        
        [ValidationMethod(ValidationCategories.Open | ValidationCategories.Save | ValidationCategories.Menu)]
        public void CheckForSingleClassInheritance(ValidationContext context)
        {
            foreach (Type childType in Types)
            {
                if (!(childType is Class))
                    continue;
                List<Type> parents = new List<Type>();
                foreach (Type parentType in childType.BaseTypes)
                {
                    if (parentType is Class) 
                        parents.Add(parentType);
                }
                if (parents.Count > 1)
                    context.LogError("Class can not inherit from multiple classes", "MultipleClassInheritance",
                                     childType);
            }
        }

        [ValidationMethod(ValidationCategories.Open | ValidationCategories.Save | ValidationCategories.Menu)]
        public void CheckForBaseClassAndInheritedClassAccessModifiers(ValidationContext context)
        {
            return;
        }

        [ValidationMethod(ValidationCategories.Open | ValidationCategories.Save | ValidationCategories.Menu)]
        public void CheckForNotImplementedAbstractMethods(ValidationContext context)
        {
            return;
        }

        [ValidationMethod(ValidationCategories.Open | ValidationCategories.Save | ValidationCategories.Menu)]
        public void CheckInheritanceForCycles(ValidationContext context)
        {
            foreach (Type type in Types)
            {
                List<Type> cycle = null;
                if ((cycle = GetCycle(type)) != null)
                {
                    context.LogError("There must be no cycles in inheritance", "InheritanceCycle", cycle.ToArray());
                    break;
                }
            }
        }

        private Dictionary<Type, Type> stack = new Dictionary<Type, Type>();
        
        private List<Type> GetCycle(Type type)
        {
            List<Type> result = null;
            if (stack.ContainsKey(type))
            {
                result = new List<Type>();
                result.AddRange(stack.Keys);
                return result;
            }
            stack.Add(type, type);
            foreach (Type baseType in type.BaseTypes)
            {
                result = GetCycle(baseType);
                if (result != null)
                    break;
            }
            stack.Remove(type);
            return result;
        }
    }
}
