using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using Microsoft.VisualStudio.Modeling.Validation;

namespace ImTryin.StateMachineLanguage
{
    [ValidationState(ValidationState.Enabled)]
    public partial class StateMachine
    {
        [ValidationMethod(ValidationCategories.Open | ValidationCategories.Save | ValidationCategories.Menu)]
        public void CheckForOnlyInitialState(ValidationContext context)
        {
            List<BaseState> initialStates = new List<BaseState>();
            foreach (BaseState state in States)
            {
                if (state is InitialState)
                    initialStates.Add(state);
            }
            if (initialStates.Count == 0)
                context.LogError("State machine has no initial states", "NoInitialStates", this);
            else if (initialStates.Count > 1)
                context.LogError("State machine has multiple initial states", "MultipleInitialStates", initialStates.ToArray());
        }

        [ValidationMethod(ValidationCategories.Open | ValidationCategories.Save | ValidationCategories.Menu)]
        public void CheckForFinalStateExistence(ValidationContext context)
        {
            bool finalState = false;
            foreach (BaseState state in States)
            {
                if (state is FinalState)
                {
                    finalState = true;
                    break;
                }
            }
            if (!finalState)
                context.LogError("State machine has no final state", "NoFinalStates", this);
        }

        [ValidationMethod(ValidationCategories.Open | ValidationCategories.Save | ValidationCategories.Menu)]
        public void CheckForUnreachableStates(ValidationContext context)
        {
            InitialState initialState = null;
            Dictionary<BaseState, BaseState> unreachableStates = new Dictionary<BaseState, BaseState>();
            foreach (BaseState state in States)
            {
                if (state is InitialState)
                    initialState = (InitialState) state;
                else
                    unreachableStates[state] = state;
            }
            if (initialState == null)
                return;

            Queue<BaseState> queue = new Queue<BaseState>();
            queue.Enqueue(initialState);
            while (queue.Count > 0)
            {
                BaseState state = queue.Dequeue();
                foreach (Transition transition in Transition.GetLinksToTargets(state))
                {
                    if (unreachableStates.ContainsKey(transition.Target))
                    {
                        queue.Enqueue(transition.Target);
                        unreachableStates.Remove(transition.Target);
                    }
                }
            }
            if (unreachableStates.Count > 0)
            {
                List<BaseState> states = new List<BaseState>();
                states.AddRange(unreachableStates.Keys);
                context.LogError("There are unreachable states in the state machine", "UnreachableStates",
                                 states.ToArray());
            }
        }

        [ValidationMethod(ValidationCategories.Open | ValidationCategories.Save | ValidationCategories.Menu)]
        public void CheckForUniqueStateNames(ValidationContext context)
        {
            Dictionary<string, List<BaseState>> statesByName = new Dictionary<string, List<BaseState>>();
            foreach (BaseState state in States)
            {
                if (!statesByName.ContainsKey(state.Name))
                    statesByName[state.Name] = new List<BaseState>();
                statesByName[state.Name].Add(state);
            }
            foreach (string name in statesByName.Keys)
            {
                if (statesByName[name].Count > 1)
                    context.LogError("States have the same names", "SameStateNames", statesByName[name].ToArray());
            }
        }

        [ValidationMethod(ValidationCategories.Open | ValidationCategories.Save | ValidationCategories.Menu)]
        public void CheckForTransitionTriggers(ValidationContext context)
        {
            foreach (BaseState state in States)
            {
                ReadOnlyCollection<Transition> transitions = Transition.GetLinksToTargets(state);
                if (transitions == null)
                    continue;
                List<Transition> notTriggeredTransitions = new List<Transition>();
                List<Transition> triggeredTransitions = new List<Transition>();
                foreach (Transition transition in transitions)
                {
                    if (transition.Trigger == string.Empty)
                        notTriggeredTransitions.Add(transition);
                    else
                        triggeredTransitions.Add(transition);
                }
                if (!(state is InitialState) && (notTriggeredTransitions.Count > 0))
                    context.LogError("Some transitions have no trigger", "NonTriggeredTransitions",
                                     notTriggeredTransitions.ToArray());
                else if ((state is InitialState) && (triggeredTransitions.Count > 0))
                    context.LogError("Transitions from initial state must not have any triggers", "TriggeredInitialTransitions",
                                     triggeredTransitions.ToArray());
            }
        }

        [ValidationMethod(ValidationCategories.Open | ValidationCategories.Save | ValidationCategories.Menu)]
        public void CheckForTransitionGuards(ValidationContext context)
        {
            Dictionary<String, List<Transition>> inconsistentTransitionsGroups = new Dictionary<String, List<Transition>>();
            foreach (BaseState state in States)
            {
                ReadOnlyCollection<Transition> transitions = Transition.GetLinksToTargets(state);
                if (transitions == null)
                    continue;
                Dictionary<string, List<Transition>> transitionsByEvent = new Dictionary<string, List<Transition>>();
                foreach (Transition transition in transitions)
                {
                    if (!transitionsByEvent.ContainsKey(transition.Trigger))
                        transitionsByEvent[transition.Trigger] = new List<Transition>();
                    transitionsByEvent[transition.Trigger].Add(transition);
                }
                foreach (string trigger in transitionsByEvent.Keys)
                {
                    List<string> guards = new List<string>();
                    foreach (Transition transition in transitionsByEvent[trigger])
                        guards.Add(transition.Guard);
                    if (!CheckForOrthogonalConditions(guards))
                        inconsistentTransitionsGroups[trigger] = transitionsByEvent[trigger];
                }
            }
            if (inconsistentTransitionsGroups.Count > 0)
            {
                foreach (string trigger in inconsistentTransitionsGroups.Keys)
                    context.LogError("State machine has a set of inconsistent transitions", "InconsistentTransitions",
                                     inconsistentTransitionsGroups[trigger].ToArray());
            }
        }

        private static bool CheckForOrthogonalConditions(List<string> conditions)
        {
            if (conditions == null) 
                return true;
            bool hasEmptyCondition = false;
            foreach (string condition in conditions)
            {
                if (condition == string.Empty)
                {
                    hasEmptyCondition = true;
                    break;
                }
            }
            if (hasEmptyCondition && (conditions.Count > 1))
                return false;
            return true;
        }
    }
}
